//: Playground - noun: a place where people can play

import UIKit

var str = "Hello, playground"

// MARK: - Node

public class Node<T> {
    typealias NodeType = Node<T>                            // just to shorten definitions
    var value: T                                            // that's the object we want to keep in the list
    var next: NodeType?                                     // since this is a class, that's the pointer to the next node
    weak var previous: NodeType?                            // and thats the one the previous node

    init(value: T) {                                        // let's initialise our node
        self.value = value                                  // remember its value
    }
}

// MARK: - LinkedList

public class LinkedList<T> {                                // the linked list, finally
    
    typealias NodeType = Node<T>                            // the type of our nodes
    fileprivate var head: NodeType?                         // the first element of our list
    private var tail: NodeType?                             // and the last one; both may be nil
    
    public fileprivate(set) var count: Int = 0              // the number of elements in the list at any given time
    
    public var isEmpty: Bool {                              // convenience property to test for emptiness
        return head == nil                                  // if the head is nil the list is empty
    }
    
    public var first: Node<T>? {                            // get the head of the list
        return head
    }
    
    public var last: Node<T>? {                             // and its tail
        return tail
    }
    
    public func append(value: T) {                          // append a new node at the end of the list
        let newNode = Node(value: value)                    // create the node with the object
        if let tailNode = tail {                            // we do have an end (list is not empty)
            newNode.previous = tailNode                     // remember that node in our new one
            tailNode.next = newNode                         // and make the new node the new tail
        } else {                                            // the list is empty
            head = newNode                                  // make our new node the head of the list
        }
        tail = newNode                                      // our new node is the (new) end of the list
        count += 1                                          // and increase our counter
    }
    
    public func nodeAt(index: Int) -> Node<T>? {            // get the node at the specified index
        if index >= 0 {                                     // make sure the index is positive
            var node = head                                 // start at the head of the list
            var i = index                                   // the count down
            while node != nil {                             // iterate over the list
                if i == 0 { return node }                   // until we finished the countdown and return the node
                i -= 1                                      // countdown
                node = node!.next                           // and take the next item
            }
        }
        return nil                                          // index is out of range
    }
    
    public func removeAll() {                               // empty the list
        head = nil                                          // clear its head
        tail = nil                                          // and its tail
        count = 0                                           // and reset the counter
    }
    
    public func remove(node: Node<T>) -> T {                // remove a node from the list
        let prev = node.previous                            // first get its neighbours
        let next = node.next
        
        if let prev = prev {                                // we are not at the top of the list
            prev.next = next                                // make the previous node point to our following one
        } else {                                            // we are at the top of the list
            head = next                                     // make our follower the new head of the list
        }
        next?.previous = prev                               // and make our follower point to our previous node
        
        if next == nil {                                    // we are at the end of the list and do not have a follower
            tail = prev                                     // make the previous node the new tail of the list
        }
        
        node.previous = nil                                 // and clear the pointers in our node
        node.next = nil
        count -= 1                                          // decrease the counter
        
        return node.value                                   // and return our object
    }
    
    private func nodesBeforeAndAfter(index: Int) -> (NodeType?, NodeType?) {    // get the nodes before and after an index
        assert(index >= 0)                                  // make sure the index is positive (+ 0)
        
        var i = index                                       // let's countdown
        var next = head                                     // start at the top of the list
        var prev: NodeType?                                 // in the beginning the previous node is nil
        
        while next != nil && i > 0 {                        // iterate over the list
            i -= 1                                          // and count down
            prev = next                                     // the previous node becomes the current one
            next = next!.next                               // and the current one the next
        }
        assert(i == 0)                                      // make sure we hit the specified index
        
        return (prev, next)                                 // and return the previous and the current node
    }
    
    private func insert(value: T, prev: NodeType?, next: NodeType?) {   // insert a new node between neighbours
        let newNode = Node(value: value)                    // create the new node from the object
        newNode.previous = prev                             // let's point to the old neighbours
        newNode.next = next
        prev?.next = newNode                                // and from the neighbours to our new node
        next?.previous = newNode
        
        if prev == nil {                                    // we inserted at the top of the list
            head = newNode                                  // the head will be our new node
        }
        if next == nil {                                    // we inserted at the end of the list
            tail = newNode                                  // the tail will be our new node
        }
        count += 1                                          // and increase the counter
    }
    
    public func insert(value: T, atIndex index: Int) {      // insert an object at a specified index
        let (prev, next) = nodesBeforeAndAfter(index: index)    // get the neighbours at the index
        insert(value: value, prev: prev, next: next)        // and insert it in between
    }
}

extension LinkedList: CustomStringConvertible {
    public var description: String {
        var text = "["
        var node = head
        while node != nil {
            text += "\(node!.value)"
            node = node!.next
            if node != nil { text += ", " }
        }
        return text + "]"
    }
}

let myFirstList = LinkedList<String>()
myFirstList.append(value: "First item")
myFirstList.append(value: "Second item")
myFirstList.append(value: "Third item")

Swift.print("The list: \(myFirstList)")
Swift.print("count: \(myFirstList.count)")

myFirstList.insert(value: "New second item", atIndex: 1)

Swift.print("The list: \(myFirstList)")
Swift.print("count: \(myFirstList.count)")

Swift.print("Remove new third item: \(myFirstList.remove(node: myFirstList.nodeAt(index: 2)!))")

Swift.print("The list: \(myFirstList)")
Swift.print("count: \(myFirstList.count)")

